<?php
/**
 * @package     CSVI
 * @subpackage  HikaShop
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - @year@ RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

namespace hikashop\com_hikashop\model\import;

use Matrix\Exception;

defined('_JEXEC') or die;

/**
 * Main processor for handling product details.
 *
 * @package     CSVI
 * @subpackage  HikaShop
 * @since       6.0
 */
class Product extends \RantaiImportEngine
{
	/**
	 * The product table
	 *
	 * @var    \HikashopTableProduct
	 * @since  6.0
	 */
	private $productTable;

	/**
	 * The product table
	 *
	 * @var    \HikashopTableProductCategory
	 * @since  6.0
	 */
	private $productCategoryTable;

	/**
	 * The media table
	 *
	 * @var    \HikashopTableFile
	 * @since  6.0
	 */
	private $fileTable;

	/**
	 * The product price table
	 *
	 * @var    \HikashopTablePrice
	 * @since  6.0
	 */
	private $priceTable;

	/**
	 * The product relations table
	 *
	 * @var    \HikashopTableProductRelated
	 * @since  6.0
	 */
	private $productRelatedTable;

	/**
	 * The HikaShop helper
	 *
	 * @var    \Com_HikashopHelperCom_Hikashop
	 * @since  6.0
	 */
	protected $helper;

	/**
	 * The HikaShop config helper
	 *
	 * @var    \Com_HikashopHelperCom_Hikashop_Config
	 * @since  6.0
	 */
	protected $helperconfig;

	/**
	 * Transliterate class
	 *
	 * @var    \CsviHelperTranslit
	 * @since  6.0
	 */
	protected $translit;

	/**
	 * List of available custom fields
	 *
	 * @var    array
	 * @since  6.6.0
	 */
	private $customFields = array();

	/**
	 * Get the list of parent products
	 *
	 * @var    array
	 * @since  7.7.0
	 */
	private $parentProducts = array();

	/**
	 * Get the list of child products with parent
	 *
	 * @var    array
	 * @since  7.7.0
	 */
	private $childWithParent = array();

	/**
	 * Keep track of which products we have deleted variants for
	 *
	 * @var    array
	 * @since  7.7.0
	 */
	private $deletedVariantProducts = array();

	/**
	 * Run this before we start.
	 *
	 * @return  void.
	 *
	 * @since   6.0
	 *
	 * @throws  \RuntimeException
	 */
	public function onBeforeStart()
	{
		// Load the transliterate class
		$this->translit = new \CsviHelperTranslit($this->template);

		// Load the custom fields
		$this->loadCustomFields();
	}

	/**
	 * Here starts the processing.
	 *
	 * @return  bool  Returns true on success | False on failure.
	 *
	 * @since   6.0.0
	 *
	 * @throws  \RuntimeException
	 */
	public function getStart()
	{
		// Process data
		foreach ($this->fields->getData() as $fields)
		{
			foreach ($fields as $name => $details)
			{
				$value = $details->value;

				// Check if the field needs extra treatment
				switch ($name)
				{
					case 'product_sale_start':
					case 'product_sale_end':
						$this->setState($name, $this->convertDate($value, 'unix'));
						break;
					case 'product_msrp':
						// Cannot clean price otherwise we lose calculations
						$this->setState($name, $this->toPeriod($value));
						break;
					case 'product_price':
						// Cannot clean price otherwise we lose calculations
						$this->setState($name, $this->toPeriod($value));
						break;
					case 'product_weight':
					case 'product_length':
					case 'product_width':
					case 'product_height':
						$this->setState($name, $this->toPeriod($value));
						break;
					case 'related_products':
						if (substr($value, -1, 1) === "|")
						{
							$value = substr($value, 0, -1);
						}

						$this->setState($name, $value);
						break;
					case 'category_id':
					case 'category_path':
						if (strlen(trim($value)) > 0)
						{
							if (strpos($value, '|') > 0)
							{
								$category_ids[$name] = explode("|", $value);
							}
							else
							{
								$category_ids[$name][] = $value;
							}

							$this->setState('category_ids', $category_ids);
						}

						$this->setState($name, $value);
						break;
					case 'product_published':
						switch ($value)
						{
							case 'n':
							case 'no':
							case 'N':
							case 'NO':
							case '0':
								$value = 0;
								break;
							default:
								$value = 1;
								break;
						}

						$this->setState($name, $value);
						break;
					case 'product_tax_category':
						if ($value === '*')
						{
							$taxId = 0;
						}
						else
						{
							$taxId = $this->getTaxCategoryId($value);

							// Check if there is any category with the given name
							if (empty($taxId))
							{
								$this->log->add('No tax category with name ' . $value . ' found');
							}
						}

						$this->setState('product_tax_id', $taxId);
						break;
					case 'default_variant':
						switch ($value)
						{
							case 'Y':
							case 'Yes':
							case 'YES':
							case 'y':
							case '1':
								$value = 1;
								break;
							default:
								$value = 0;
								break;
						}

						$this->setState($name, $value);
						break;
					case 'warehouse_name':
						$wareHouseId = $this->getWarehouseId($value);
						$this->setState('product_warehouse_id', $wareHouseId);
						break;
					default:
						$this->setState($name, $value);
						break;
				}
			}
		}

		$field = $this->template->get('update_based_on', 'product_code');

		if ($this->getState($field, false))
		{
			$this->productTable->setKeyName($field);
		}
		else
		{
			$field = 'product_id';
		}

		if ($this->getState($field, false))
		{
			// Set the record identifier
			$this->recordIdentity = $this->getState($field, '');

			// Load the current product data
			$this->loaded = true;

			if ($this->productTable->load($this->getState($field, false)))
			{
				if (!$this->template->get('overwrite_existing_data'))
				{
					$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_DATA_EXISTS_PRODUCT_SKU', $this->getState($field, '')));
					$this->loaded = false;
				}
				else
				{
					$this->setState('product_id', $this->productTable->get('product_id', 0));
				}
			}
			else
			{
				$product_id = $this->getState('product_id', false);

				if (!$this->getState($field, false) && empty($product_id))
				{
					$this->log->addStats('incorrect', 'COM_CSVI_DEBUG_NO_SKU');
					$this->log->add('COM_CSVI_DEBUG_NO_SKU');

					$this->loaded = false;
				}
				else
				{
					// Product is not found so we need to reset to the primary key field
					$this->productTable->setKeyName('product_id');
					$this->log->add('Process SKU: ' . $this->recordIdentity, false);
				}
			}

			// Check for child product and get parent SKU if it is
			if ($this->getState('product_parent_code', false))
			{
				$this->productParentCode();
			}
		}
		else
		{
			$this->loaded = false;

			$field = $this->template->get('update_based_on', 'product_code');
			$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_MISSING_REQUIRED_FIELDS', $field), 'product');
		}

		return true;
	}

	/**
	 * Get the tax id of the given tax category name
	 *
	 * @param   string $taxCategoryName The name of tax category.
	 *
	 * @return  int  Returns tax id corresponding to tax category name.
	 *
	 * @since   6.4.0
	 *
	 * @throws  \RuntimeException
	 */
	private function getTaxCategoryId($taxCategoryName)
	{
		$query = $this->db->getquery(true)
			->select($this->db->quoteName('category_id'))
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quotename('category_type') . ' = ' . $this->db->quote('tax'))
			->where($this->db->quotename('category_name') . ' = ' . $this->db->quote($taxCategoryName));
		$this->db->setQuery($query);

		return $this->db->loadResult();
	}

	/**
	 * Process each record and store it in the database.
	 *
	 * @return  bool  Returns true if all is OK | Returns false if no product SKU or product ID can be found.
	 *
	 * @since   6.0
	 *
	 * @throws  \RuntimeException
	 * @throws  \InvalidArgumentException
	 */
	public function getProcessRecord()
	{
		if ($this->loaded)
		{
			// Get the needed data
			$product_id     = $this->getState('product_id', false);
			$product_delete = $this->getState('product_delete', 'N');

			// User wants to delete the product
			if ($product_id && $product_delete === 'Y')
			{
				$this->deleteProduct();
			}
			elseif (!$product_id && $product_delete === 'Y')
			{
				$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_NO_PRODUCT_ID_NO_DELETE', $this->recordIdentity));
			}
			elseif (!$product_id && $this->template->get('ignore_non_exist'))
			{
				// Do nothing for new products when user chooses to ignore new products
				$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_DATA_EXISTS_IGNORE_NEW', $this->recordIdentity));
			}
			else
			{
				// Find the manufacturer ID
				if ($this->getState('brand', false))
				{
					$this->setState('product_manufacturer_id', $this->helper->getManufacturerId($this->getState('brand')));
				}

				// Process the ICEcat product features
				$this->setState('features', $this->fields->get('features'));

				if ($this->template->get('use_icecat', false, 'bool') && $this->getState('features', false))
				{
					$this->icecatFeatures();
				}

				// Handle the access level(s)
				if ($accessIds = $this->getState('product_access', false))
				{
					if ($accessIds !== 'all')
					{
						$accessIds = $this->helper->processAccessLevel('product_access');
					}

					$this->setState('product_access', $accessIds);
				}

				if (!$this->getState('product_url', false))
				{
					$this->setState('product_url', '');
				}

				if (!$this->getState('product_keywords', false))
				{
					$this->setState('product_keywords', '');
				}

				if (!$this->getState('product_meta_description', false))
				{
					$this->setState('product_meta_description', '');
				}

				if (!$this->getState('product_average_score', false))
				{
					$this->setState('product_average_score', 0);
				}

				// Process product info
				if ($this->productQuery())
				{
					// Handle the images
					$this->processMedia();

					// Check if the price is to be updated
					if ($this->getState('product_price', false) || $this->getState('price_with_tax', false) || $this->getState('price_access_new'))
					{
						$this->priceQuery();
					}

					/**
					 * Process related products/categories
					 * Related products are first input in the database as SKU
					 * At the end of the import, this is converted to product ID
					 */
					if ($this->getState('related_products'))
					{
						$this->processRelatedProducts();
					}

					if ($this->getState('product_options'))
					{
						$this->processProductOptions();
					}

					// Process category path
					if ($this->getState('category_path', false) || $this->getState('category_id', false))
					{
						if ($this->getState('category_ids', false) || $this->getState('category_id', false))
						{
							require_once JPATH_PLUGINS . '/csviaddon/hikashop/com_hikashop/helper/com_hikashop_category_product.php';
							$categoryHelper = new \Com_HikashopHelperCom_Hikashop_Category_Product($this->db, $this->template, $this->log);
							$areas          = $this->getState('category_ids');

							if ($areas)
							{
								// Delete old entries only if the user wants us to, set in template config append_categories
								if (!$this->template->get('append_categories', false))
								{
									$query = $this->db->getQuery(true)
										->delete($this->db->quoteName('#__hikashop_product_category'))
										->where(
											$this->db->quoteName('product_id') . ' = ' . (int) $this->getState('product_id')
										);
									$this->db->setQuery($query)->execute();
									$this->log->add('Delete old category references');
								}
								else
								{
									$this->log->add('Do not delete old category references, going to append the categories', false);
								}

								// Check the categories
								if (array_key_exists('category_id', $areas))
								{
									foreach ($areas['category_id'] as $order => $categoryId)
									{
										$category_product_values = array();

										// Insert new product/category relationships
										$category_product_values['product_id']  = $this->getState('product_id');
										$category_product_values['category_id'] = $categoryId;
										$category_product_values['ordering']    = $order + 1;
										$this->productCategoryTable->bind($category_product_values);
										$this->productCategoryTable->store();
										$this->productCategoryTable->reset();
									}
								}
								elseif (array_key_exists('category_path', $areas))
								{
									foreach ($areas['category_path'] as $order => $path)
									{
										$category_product_values = array();
										$this->log->add('Process path: ' . $path, false);

										$categoryId = $categoryHelper->getCategoryIdByPath($path);

										// Insert new product/category relationships
										$category_product_values['product_id']  = $this->getState('product_id');
										$category_product_values['category_id'] = $categoryId;
										$category_product_values['ordering']    = $order + 1;
										$this->productCategoryTable->bind($category_product_values);
										$this->productCategoryTable->store();
										$this->productCategoryTable->reset();
									}
								}
							}
						}
					}

					// Check if the field is a custom field used as an available field
					$this->processCustomAvailableFields();

					// Handle multiple language text
					$this->helper->processMultiLanguageFields($this->state, $this->productTable, 'product');
				}
			}

			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * Execute any processes to finalize the import.
	 *
	 * @param   array $fields list of fields used for import
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function getPostProcessing($fields = array())
	{
		// Related products
		if (in_array('related_products', $fields))
		{
			$this->postProcessRelatedProducts();
		}
	}

	/**
	 * Load the product related tables.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function loadTables()
	{
		$this->productTable         = $this->getTable('Product');
		$this->fileTable            = $this->getTable('File');
		$this->priceTable           = $this->getTable('Price');
		$this->productRelatedTable  = $this->getTable('Productrelated');
		$this->productCategoryTable = $this->getTable('Productcategory');
	}

	/**
	 * Cleaning the product related tables.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function clearTables()
	{
		$this->productTable->reset();
		$this->fileTable->reset();
		$this->priceTable->reset();
		$this->productRelatedTable->reset();
	}

	/**
	 * Get a list of custom fields that can be used as available field.
	 *
	 * @return  void.
	 *
	 * @since   6.6.0
	 *
	 * @throws  \RuntimeException
	 */
	private function loadCustomFields()
	{
		$query = $this->db->getQuery(true)
			->select(
				$this->db->quoteName(
					array(
						'characteristic_id',
						'characteristic_alias',
					)
				)
			)
			->from($this->db->quoteName('#__hikashop_characteristic'))
			->where($this->db->quoteName('characteristic_parent_id') . ' = 0');
		$this->db->setQuery($query);
		$this->customFields = $this->db->loadObjectList();

		$this->log->add('Load the custom fields');
	}

	/**
	 * Creates either an update or insert SQL query for a product.
	 *
	 * @return  bool true if the query executed successful|false if the query failed.
	 *
	 * @since   3.0
	 *
	 * @throws  \InvalidArgumentException
	 */
	private function productQuery()
	{
		// Set the modified date as we are modifying the product
		if (!$this->getState('modified_on'))
		{
			$this->setState('product_modified', $this->date->toUnix());
		}

		// Add a creating date if there is no product_id
		if (!$this->getState('product_id'))
		{
			// A new product, make sure there is a category if it is a parent product
			if (!$this->getState('category_path', false)
				&& !$this->getState('category_id', false)
				&& !$this->getState('childProduct', false))
			{
				// No category info imported, force the product root category
				$this->setState('category_id', 2);
			}

			$this->setState('product_created', $this->date->toUnix());

			// Process default values
			$defaults = array(
				'product_weight',
				'product_published',
				'product_weight_unit',
				'product_dimension_unit',
				'product_type',
			);

			foreach ($defaults as $field)
			{
				switch ($field)
				{
					case 'product_weight':
					case 'product_published':
						if (!$this->getState($field, false))
						{
							$this->setState($field, 0);
						}
						break;
					case 'product_weight_unit':
						if (!$this->getState($field, false))
						{
							$this->setState($field, 'kg');
						}
						break;
					case 'product_dimension_unit':
						if (!$this->getState($field, false))
						{
							$this->setState($field, 'm');
						}
						break;
					case 'product_type':
						if (!$this->getState($field, false))
						{
							$this->setState('product_type', 'main');
						}
						break;
					case 'product_access':
						if (!$this->getState($field, false))
						{
							$this->setState('product_access', 'all');
						}
						break;
				}
			}
		}

		// If there is product name set, assign it to table field to generate alias
		if ($this->getState('product_name', false))
		{
			$this->productTable->product_name = $this->getState('product_name', false);
		}

		if ($this->getState('product_id', false))
		{
			$this->productTable->product_id = $this->getState('product_id', false);
		}

		// If there is product alias set, assign it to table field to generate alias
		if ($this->getState('product_alias', false))
		{
			$this->productTable->product_alias = $this->getState('product_alias', false);
		}

		// Recreate the alias if user wants to or if no alias set in import file
		if ($this->template->get('recreate_alias', false) || !$this->getState('product_alias', false))
		{
			$this->productTable->createProductAlias();
			$this->setState('product_alias', $this->productTable->product_alias);
		}

		// Store the product
		if (!$this->productTable->save($this->state))
		{
			return false;
		}

		// Set the product ID
		$this->setState('product_id', $this->productTable->get('product_id'));

		// All good
		return true;
	}

	/**
	 * Process related products.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 *
	 * @throws  \RuntimeException
	 */
	private function processRelatedProducts()
	{
		$relatedProducts = explode('|', $this->getState('related_products'));

		$query   = 'INSERT IGNORE INTO ';
		$query   .= $this->db->quoteName('#__csvi_related_products') . ' VALUES ';
		$entries = array();

		foreach ($relatedProducts as $relatedproduct)
		{
			$entries[] = '(' . $this->db->quote($this->getState('product_code')) . ', ' . $this->db->quote($relatedproduct) . ')';
		}

		$query .= implode(',', $entries);
		$this->db->setQuery($query)->execute();

		// Remove any existing product relations
		$this->productRelatedTable->deleteRelated($this->getState('product_id'), 'related');
	}

	/**
	 * Post-process related products.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 *
	 * @throws  \RuntimeException
	 */
	private function postProcessRelatedProducts()
	{
		// Get the related products
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('p1.product_id', 'product_id') . ',' . $this->db->quoteName('p2.product_id', 'product_related_id'))
			->from($this->db->quoteName('#__csvi_related_products', 'r'))
			->leftJoin(
				$this->db->quoteName('#__hikashop_product', 'p1')
				. ' ON ' . $this->db->quoteName('r.product_sku') . ' = ' . $this->db->quoteName('p1.product_code')
			)
			->leftJoin(
				$this->db->quoteName('#__hikashop_product', 'p2')
				. ' ON ' . $this->db->quoteName('r.related_sku') . ' = ' . $this->db->quoteName('p2.product_code')
			);
		$this->db->setQuery($query);
		$this->log->add('Process related products', true);
		$relations = $this->db->loadObjectList();

		if (!empty($relations))
		{
			// Store the new relations
			foreach ($relations as $ordering => $related)
			{
				// Insert or Updated only when there is a value
				if ($related->product_related_id > 0 && $related->product_related_id !== '')
				{
					// Build the object to store
					$related->product_related_type     = 'related';
					$related->product_related_ordering = $ordering + 1;

					// Bind the data
					$this->productRelatedTable->save($related);

					// Clean the table object for next insert
					$this->productRelatedTable->reset();
				}
			}

			// Empty the relations table
			$this->db->truncateTable('#__csvi_related_products');
		}
		else
		{
			$this->log->addStats('incorrect', 'COM_CSVI_NO_RELATED_PRODUCTS_FOUND');
		}
	}

	/**
	 * Process media files.
	 *
	 * @return  bool Returns true on OK | False on failure.
	 *
	 * @since   4.0
	 *
	 * @throws  \RuntimeException
	 */
	private function processMedia()
	{
		$generate_image = $this->template->get('auto_generate_image_name', false);
		$file_path      = $this->getState('file_path', false);

		// Check if any image handling needs to be done
		if ($file_path || $generate_image)
		{
			// Check if we have any images
			if (!$file_path && $generate_image)
			{
				$this->createImageName();
			}

			// Create an array of images to process
			$images       = explode('|', $file_path);
			$titles       = explode('|', $this->getState('file_name'));
			$descriptions = explode('|', $this->getState('file_description'));
			$ordering     = explode('|', $this->getState('file_ordering', '1'));
			$max_width    = $this->template->get('resize_max_width', 1024);
			$max_height   = $this->template->get('resize_max_height', 768);

			// Image handling
			$imageHelper = new \CsviHelperImage($this->template, $this->log, $this->csvihelper);

			// Delete existing image links
			if ($this->template->get('delete_product_images', false))
			{
				$query = $this->db->getQuery(true)
					->delete($this->db->quoteName('#__hikashop_file'))
					->where($this->db->quoteName('file_type') . '=' . $this->db->quote('product'))
					->where($this->db->quoteName('file_ref_id') . '=' . (int) $this->getState('product_id'));
				$this->db->setQuery($query)->execute();
				$this->log->add('Delete images');
			}

			foreach ($images as $key => $image)
			{
				$image = trim($image);

				// Create image name if needed
				$img_counter = $key + 1;

				if (count($images) === 1)
				{
					$img_counter = 0;
				}

				if ($generate_image)
				{
					$name = $this->createImageName($img_counter);

					if (!empty($name))
					{
						$image = $name;
					}
				}

				if (!empty($image))
				{
					// Get the image path from the HikaShop config
					$configPath = $this->helperconfig->get('uploadfolder');
					$imgpath    = $this->template->get('file_location_product_files', $configPath);

					if (substr($imgpath, -1) !== '/')
					{
						$imgpath .= '/';
					}

					// Verify the original image
					if ($imageHelper->isRemote($image))
					{
						$original  = $image;
						$full_path = $imgpath;
					}
					else
					{
						// Check if the image contains the image path
						$dirname = dirname($image) . '/';

						if (strpos($dirname, $imgpath) !== false)
						{
							$imageLeftPath = str_replace($imgpath, '', $dirname);
							$image         = basename($image);

							if ($imageLeftPath)
							{
								$image = $imageLeftPath . $image;
							}
						}

						$original = $imgpath . $image;

						// Get subfolders
						$path_parts = pathinfo($original);
						$full_path  = $path_parts['dirname'] . '/';
					}

					// Generate image names
					if ($this->template->get('process_image', false))
					{
						if ($generate_image)
						{
							$file_details = $imageHelper->processImage($original, $full_path, $this->getState('product_full_image_output'));
						}
						else
						{
							$file_details = $imageHelper->processImage($original, $full_path, array_key_exists($key, $titles) ? $titles[$key] : '');
						}
					}
					else
					{
						$file_details['exists']      = true;
						$file_details['isimage']     = $imageHelper->isImage(JPATH_SITE . '/' . $image);
						$file_details['name']        = $image;
						$file_details['output_name'] = basename($image);
						$file_details['output_path'] = $full_path;
					}

					// Process the file details
					if ($file_details['exists'])
					{
						$media = array();

						if ($this->template->get('autofill'))
						{
							$media['file_name']        = $file_details['output_name'];
							$media['file_description'] = $file_details['output_name'];
						}
						else
						{
							$media['file_name']        = array_key_exists($key, $titles) ? $titles[$key] : $file_details['output_name'];
							$media['file_description'] = array_key_exists($key, $descriptions) ? $descriptions[$key] : '';
						}

						$imageLeftPath      = str_replace($configPath, '', $file_details['output_path']);
						$media['file_path'] = (!$imageLeftPath) ? $file_details['output_name'] : $imageLeftPath . $file_details['output_name'];

						// If its an external image and not saving on server store the external URL
						if (0 === strpos($file_details['name'], 'http') && !$this->template->get('save_images_on_server', false))
						{
							$media['file_path'] = $file_details['output_path'] . $file_details['output_name'];
						}

						$media['file_type']          = 'product';
						$media['file_ref_id']        = $this->getState('product_id');
						$media['file_free_download'] = 0;
						$media['file_ordering']      = $ordering[$key] ?? '';
						$media['file_limit']         = 0;

						// Create the thumbnail
						if ($file_details['isimage'] && $this->template->get('thumb_create'))
						{
							$mainfile = JPATH_SITE . '/' . $file_details['output_path'] . $file_details['output_name'];
							$this->log->add('Reading file ' . $mainfile . ' for image size', false);

							if (file_exists($mainfile))
							{
								$thumb_sizes = getimagesize($mainfile);

								if ($thumb_sizes[0] < $max_width || $thumb_sizes[1] < $max_height)
								{
									$destination = $imgpath . 'thumbnail_' . $this->template->get('thumb_height') . 'x' . $this->template->get('thumb_width') . '/';
									$imageHelper->createThumbnail($file_details['output_path'] . $file_details['output_name'], $destination, $file_details['output_name']);
								}
							}
						}

						// Bind the media data
						$this->fileTable->bind($media);

						// Check if the media image already exists
						$this->fileTable->check();

						// Store the media data
						if ($this->fileTable->store())
						{
							// Watermark the image
							if ($this->template->get('full_watermark', 'image') && $file_details['isimage'])
							{
								$imageHelper->addWatermark(JPATH_SITE . '/' . $media['file_path']);
							}
						}
						else
						{
							$this->log->addStats('incorrect', \JText::sprintf('COM_CSVI_MEDIA_NOT_ADDED', $this->fileTable->getError()));

							return false;
						}

						// Reset the product media table
						$this->fileTable->reset();
					}
				}
			}
		}

		return true;
	}

	/**
	 * Creates either an update or insert SQL query for a product price.
	 *
	 * @return  true or false.
	 *
	 * @since   6.0
	 *
	 * @throws  \RuntimeException
	 */
	private function priceQuery()
	{
		// Check if the price is including or excluding tax
		if ($this->getState('product_tax', false) && $this->getState('price_with_tax', false) && is_null($this->getState('product_price')))
		{
			if (!$this->getState('price_with_tax', false))
			{
				$this->setState('product_price', null);
			}
			else
			{
				$this->setState('product_price', ($this->getState('price_with_tax') / (1 + ($this->getState('product_tax') / 100))));
			}
		}
		elseif (strlen($this->getState('product_price', '')) == 0)
		{
			$this->setState('product_price', null);
		}

		// Get the price access
		$priceAccess      = $this->getState('price_access', false);
		$priceAccessLevel = 'all';

		// Validate the price access level
		if ($priceAccess)
		{
			// Check if we need to change the access
			$priceAccessLevel = $this->helper->processAccessLevel('price_access');

			if ($priceAccessLevel)
			{
				$this->setState('price_access', $priceAccessLevel);
			}
			else
			{
				$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_PRICE_ACCESSGROUP_NOT_EXIST', $priceAccess));

				return false;
			}
		}

		// Bind some of the fields
		$this->setState('price_product_id', $this->getState('product_id'));
		$this->setState('price_value', $this->getState('product_price'));

		// Bind the fields to check for an existing price
		$this->priceTable->bind($this->state);

		$this->log->add('Access level price: ' . ($priceAccess) ? $priceAccess : 'not set', false);

		// Check if the price already exists
		if (!$this->priceTable->check())
		{
			// Price doesn't exist
			if (!$this->priceTable->get('price_site_id'))
			{
				$this->priceTable->set('price_site_id', '');
			}

			// Check if we need to assign the all shopper group, if no shopper group has been set
			$this->setState('price_access', $priceAccessLevel ?: 'all');
		}
		else
		{
			// Price exist, see if we need to assign a new usergroup
			if ($price_access_new = $this->getState('price_access_new', false))
			{
				if ($price_access_new !== 'all')
				{
					// Check if we need to change the access
					$price_access_new = $this->helper->processAccessLevel('price_access_new');
				}

				if ($price_access_new)
				{
					$this->setState('price_access', $price_access_new);
				}
				else
				{
					$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_PRICE_ACCESSGROUP_NOT_EXIST', $this->getState('price_access_new', false)));

					return false;
				}
			}
		}

		// Check for a currency code if needed
		if ($this->getState('currency_code', false))
		{
			$this->setState('price_currency_id', $this->helper->getCurrencyId($this->getState('currency_code', false)));
		}

		// Check if we need to set a currency
		if (!$this->getState('currency_code', false) && !$this->getState('price_currency_id', false))
		{
			$this->setState('price_currency_id', $this->helperconfig->get('main_currency'));
		}

		// Re-bind because the values are loaded during the check
		$this->priceTable->bind($this->state);

		// Store the price
		$this->priceTable->store();

		return true;
	}

	/**
	 * Create image name.
	 *
	 * Check if the user wants to have CSVI VirtueMart create the image names if so
	 * create the image names without path.
	 *
	 * @param   int $ordering The number to apply to a generated image name.
	 *
	 * @return  string  The name of the image.
	 *
	 * @since   3.0
	 */
	private function createImageName($ordering = 0)
	{
		$this->log->add('Create image name');

		// Create extension
		$ext = $this->template->get('autogenerateext');

		// Check if the user wants to convert the images to a different type
		switch ($this->template->get('type_generate_image_name'))
		{
			case 'product_code':
				$this->log->add('Create image based on product SKU');

				if ($this->getState('product_code', false))
				{
					$name = $this->getState('product_code');
				}
				else
				{
					$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_SKU');

					return false;
				}
				break;
			case 'product_name':
				$this->log->add('Create image based on product name');

				if ($this->productTable->get('product_name'))
				{
					$name = str_replace(' ', '_', $this->productTable->get('product_name'));
				}
				else
				{
					$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_NAME');

					return false;
				}
				break;
			case 'product_id':
				$this->log->add('Create image based on product ID');

				if ($this->getState('product_id', false))
				{
					$name = $this->getState('product_id');
				}
				else
				{
					$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_ID');

					return false;
				}
				break;
			default:
				$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_SKU');

				return false;
		}

		// Build the new name
		$image_name = $name . '.' . $ext;

		if ($ordering > 0)
		{
			$image_name = $name . '_' . $ordering . '.' . $ext;
		}

		$this->log->add('Created image name ' . $image_name);
		$this->setState('product_full_image_output', $image_name);

		// Check if the user is supplying image data
		if ($ordering === 0 && !$this->getState('file_path'))
		{
			$this->setState('file_path', $this->getState('product_full_image_output'));

			return $this->getState('file_path');
		}

		return '';
	}

	/**
	 * Combine the ICEcat features.
	 *
	 * @return  void.
	 *
	 * @since   4.3
	 */
	private function icecatFeatures()
	{
		$table = '<table id="prod_features">';
		$table .= '<thead></thead>';
		$table .= '<tfoot></tfoot>';
		$table .= '<tbody>';

		foreach ($this->getState('features') as $details)
		{
			foreach ($details as $feature => $values)
			{
				$table .= '<tr><td colspan="2" class="feature">' . $feature . '</td></tr>';

				foreach ($values as $name => $value)
				{
					$table .= '<tr><td>' . $name . '</td><td>' . $value . '</td></tr>';
				}
			}
		}

		$table .= '</tbody>';
		$table .= '</table>';

		// Add the table to the product desc
		$this->setState('product_description', $this->getState('product_description') . $table);
	}

	/**
	 * Delete a product and its references.
	 *
	 * @return  bool  Always returns true.
	 *
	 * @since   4.0
	 */
	private function deleteProduct()
	{
		// Deletion based on the real primary key
		$this->productTable->setKeyName('product_id');

		// Get the product ID
		$product_id = $this->getState('product_id');

		// Delete the product
		if ($this->productTable->delete($product_id))
		{
			// Delete category reference
			$query = $this->db->getQuery(true);
			$query->delete($this->db->quoteName('#__hikashop_product_category'));
			$query->where($this->db->quoteName('product_id') . ' = ' . (int) $product_id);
			$this->db->setQuery($query)->execute();
			$this->log->add('COM_CSVI_DEBUG_DELETE_CATEGORY_XREF', true);

			// Delete related products
			$query = $this->db->getQuery(true);
			$query->delete($this->db->quoteName('#__hikashop_product_related'));
			$query->where($this->db->quoteName('product_id') . ' = ' . (int) $product_id);
			$this->db->setQuery($query)->execute();
			$this->log->add('COM_CSVI_DEBUG_DELETE_RELATED_XREF', true);

			// Delete prices
			$query = $this->db->getQuery(true);
			$query->delete($this->db->quoteName('#__hikashop_price'));
			$query->where($this->db->quoteName('price_product_id') . ' = ' . (int) $product_id);
			$this->db->setQuery($query)->execute();
			$this->log->add('COM_CSVI_DEBUG_DELETE_PRICES_XREF', true);

			// Delete media
			$query = $this->db->getQuery(true);
			$query->delete($this->db->quoteName('#__hikashop_file'));
			$query->where($this->db->quoteName('file_type') . ' IN (' . $this->db->quote('product') . ',' . $this->db->quote('file') . ')');
			$query->where($this->db->quoteName('file_ref_id') . ' = ' . (int) $product_id);
			$this->db->setQuery($query)->execute();
			$this->log->add('COM_CSVI_DEBUG_DELETE_MEDIA_XREF', true);

			// Delete votes
			$query = $this->db->getQuery(true);
			$query->delete($this->db->quoteName('#__hikashop_vote'));
			$query->where($this->db->quoteName('vote_type') . ' = ' . $this->db->quote('product'));
			$query->where($this->db->quoteName('vote_ref_id') . ' = ' . (int) $product_id);
			$this->db->setQuery($query)->execute();
			$this->log->add('COM_CSVI_DEBUG_DELETE_RATING_VOTES_XREF', true);
		}
		else
		{
			$this->log->addStats('incorrect', \JText::sprintf('COM_CSVI_PRODUCT_NOT_DELETED', $this->getState('recordIdentity')));
		}

		return true;
	}

	/**
	 * Get the product parent sku if it is a child product.
	 *
	 * @return  void.
	 *
	 * @since   6.6.0
	 *
	 * @throws  \RuntimeException
	 */
	private function productParentCode()
	{
		$this->log->add('Looking for parent code', false);

		if ($this->getState('product_code', false))
		{
			$productParentCode = $this->getState('product_parent_code', false);
			$this->setState('product_parent_id', 0);
			$this->setState('childProduct', false);

			// Check if we are dealing with a child product
			if ($productParentCode && $productParentCode !== $this->getState('product_code'))
			{
				$this->setState('childProduct', true);

				// Get the parent id first
				$query = $this->db->getQuery(true)
					->select($this->db->quoteName('product_id'))
					->from($this->db->quoteName('#__hikashop_product'))
					->where($this->db->quoteName('product_code') . ' = ' . $this->db->quote($productParentCode));
				$this->db->setQuery($query);
				$this->setState('product_parent_id', $this->db->loadResult());
				$this->log->add('Checking product parent ID');
			}
		}
	}

	/**
	 * Process custom fields that are used as available field.
	 *
	 * @return  void.
	 *
	 * @since   6.6.0
	 *
	 * @throws  \RuntimeException
	 */
	private function processCustomAvailableFields()
	{
		// Create the queries
		if ($this->customFields)
		{
			// Get the import fields to check if the custom field is been imported
			$importFields = $this->fields->getFieldnames();

			foreach ($this->customFields as $field)
			{
				if ($field->characteristic_alias && in_array($field->characteristic_alias, $importFields))
				{
					// Get the product IDs
					$variantProductId = $this->getState('product_id');
					$productParentId  = $this->getState('product_parent_id');

					// Check if parent has children
					if (in_array($productParentId, $this->parentProducts))
					{
						$this->childWithParent[] = $productParentId;
					}

					// Get the characteristic alias as that is the unique value
					$alias = $field->characteristic_alias;
					$this->log->add('Processing characteristic available field: ' . $alias, false);

					// Get the characteristic parent ID
					$characteristicParentId = $this->getCharacteristicIdByAlias($alias);

					// Insert the defaults for parent product
					if (!$this->getState('product_parent_id'))
					{
						$this->clearVariants($variantProductId);
					}

					if (!$characteristicParentId && !$this->getState($alias, false))
					{
						$this->log->add('No characterstic parent ID and no value');
						continue;
					}

					// Clean the variants table before proceeding to import
					if ($characteristicParentId && $productParentId)
					{
						$this->clearVariants($variantProductId);
					}

					// Find the characteristic ID for value
					$value = $this->getState($alias, false);

					// Only process characteristic if it has a value
					if (!$value)
					{
						continue;
					}

					$characteristicId = $this->getCharacteristicIdByValue($value, $characteristicParentId);

					// The charactaristic value doesn't exist yet, add it
					if (!$characteristicId)
					{
						$characteristicId = $this->storeCharacteristic($value, $characteristicParentId);
					}

					// Check if the characteristic exists
					$variantCharacteristicId = $this->getVariantCharacteristicId($variantProductId, $characteristicId);

					// Process the variant characteristic
					$this->log->add('Process the variant characteristic ' . $this->getState($alias) . ' for product ' . $this->getState('product_code'), false);

					if ($variantCharacteristicId)
					{
						$this->updateVariantCharacteristic($variantProductId, $characteristicId, $field->characteristic_id);
					}
					else
					{
						$this->addVariantCharacteristic($variantProductId, $characteristicId);
					}

					// Set the default variant if user wants to
					if ($this->getState('default_variant', false))
					{
						$this->storeDefaultVariant($productParentId, $characteristicId, $alias);
						$this->processCharacteristicsParentProduct($productParentId, $characteristicParentId);
					}
				}
				else
				{
					$this->log->add('Given characteristic ' . $field->characteristic_alias . ' alias is empty', false);
				}
			}
		}
		else
		{
			$this->log->add('No custom available fields found', false);
		}
	}

	/**
	 * Get the characterstic ID.
	 *
	 * @param   string  $alias  The alias to find the ID for.
	 *
	 * @return  integer  The characteristic ID.
	 *
	 * @since   7.7.0
	 */
	private function getCharacteristicIdByAlias($alias)
	{
		// Get the characteristic ID
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('characteristic_id'))
			->from($this->db->quoteName('#__hikashop_characteristic'))
			->where($this->db->quoteName('characteristic_alias') . ' = ' . $this->db->quote($alias));
		$this->db->setQuery($query);
		$this->log->add('Find the characteristic ID by alias for ' . $alias);

		return $this->db->loadResult();
	}

	/**
	 * Process the characteristics for the parent product.
	 *
	 * @param   integer  $productParentId  The parent ID
	 * @param   integer  $characteristicParentId  The parent characteristic ID
	 *
	 * @return  void
	 *
	 * @since   7.7.0
	 */
	private function processCharacteristicsParentProduct($productParentId, $characteristicParentId)
	{
		$this->parentProducts[] = $productParentId;

		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('variant_characteristic_id'))
			->from($this->db->quoteName('#__hikashop_variant'))
			->where($this->db->quoteName('variant_product_id') . ' = ' . (int) $productParentId)
			->where($this->db->quoteName('variant_characteristic_id') . ' = ' . (int) $characteristicParentId);
		$this->db->setQuery($query);

		$this->log->add('Check if the characteristic for the parent is set in the variant table');

		if (!$this->db->loadResult())
		{
			$query->clear()
				->insert($this->db->quoteName('#__hikashop_variant'))
				->columns(
					$this->db->quoteName(
						[
							'variant_characteristic_id',
							'variant_product_id',
							'ordering'
						]
					)
				)
				->values($characteristicParentId . ', ' . $productParentId . ', 1');
			$this->db->setQuery($query)->execute();
			$this->log->add('Store parent product variant for product ' . $this->getState('product_code', false));
		}
	}

	/**
	 * Get the characteristic ID.
	 *
	 * @param   string   $value                   The value to get the ID for
	 * @param   integer  $characteristicParentId  The parent characteristic ID
	 *
	 * @return  integer The characteristic ID.
	 *
	 * @since   7.7.0
	 */
	private function getCharacteristicIdByValue($value, $characteristicParentId)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('characteristic_id'))
			->from($this->db->quoteName('#__hikashop_characteristic'))
			->where($this->db->quoteName('characteristic_value') . ' = ' . $this->db->quote($value))
			->where($this->db->quoteName('characteristic_parent_id') . ' = ' . (int) $characteristicParentId);
		$this->db->setQuery($query);

		$this->log->add('Get characteristic ID by value for ' . $value);

		return $this->db->loadResult();
	}

	/**
	 * Add the characteristic value.
	 *
	 * @param   string   $value                   The value to get the ID for
	 * @param   integer  $characteristicParentId  The parent characteristic ID
	 *
	 * @return  integer  The characteristic ID
	 *
	 * @since   7.7.0
	 */
	private function storeCharacteristic($value, $characteristicParentId)
	{
		// Find out the next ordering position
		$query = $this->db->getQuery(true)
			->select('MAX(' . $this->db->quoteName('characteristic_ordering') . ') + 1 AS' . $this->db->quoteName('ordering'))
			->from($this->db->quoteName('#__hikashop_characteristic'))
			->where($this->db->quoteName('characteristic_parent_id') . ' = ' . (int) $characteristicParentId)
			->order($this->db->quoteName('characteristic_ordering'));
		$this->db->setQuery($query);
		$characteristicOrdering = $this->db->loadResult();

		$query->clear()
			->insert($this->db->quoteName('#__hikashop_characteristic'))
			->columns(
				$this->db->quoteName(
					array(
						'characteristic_parent_id',
						'characteristic_value',
						'characteristic_ordering'
					)
				)
			)
			->values($characteristicParentId . ', ' . $this->db->quote($value) . ', ' . (int) $characteristicOrdering);

		$this->db->setQuery($query)->execute();
		$this->log->add('Added characterstic ' . $value . ' for ' . $value);

		return $this->db->insertid();
	}

	/**
	 * Clear variants for a given product.
	 *
	 * @param   integer  $variantProductId  The variant product ID to delete variants for
	 *
	 * @return  void
	 *
	 * @since   7.7.0
	 */
	private function clearVariants($variantProductId)
	{
		if (in_array($variantProductId, $this->deletedVariantProducts))
		{
			return;
		}

		$query = $this->db->getQuery(true)
			->delete($this->db->quoteName('#__hikashop_variant'))
			->where($this->db->quoteName('variant_product_id') . ' = ' . (int) $variantProductId);
		$this->db->setQuery($query)->execute();
		$this->log->add('Delete existing Characteristics for a product ' . $this->getState('product_code', false) . '  before importing');

		$this->deletedVariantProducts[] = $variantProductId;
	}

	/**
	 * Get the variant characteristic.
	 *
	 * @param   integer  $variantProductId  The ID of the product ID
	 * @param   integer  $characteristicId  The ID of the characteristic ID
	 *
	 * @return  integer The variant characteristic ID.
	 *
	 * @since   7.7.0
	 */
	private function getVariantCharacteristicId($variantProductId, $characteristicId)
	{
		// Check if the characteristic exists
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('variant_characteristic_id'))
			->from($this->db->quoteName('#__hikashop_variant'))
			->where($this->db->quoteName('variant_product_id') . ' = ' . (int) $variantProductId)
			->where($this->db->quoteName('variant_characteristic_id') . ' = ' . (int) $characteristicId);
		$this->db->setQuery($query);

		$this->log->add('Check if the characteristic exists for product ' . $this->getState('product_code', false));

		return $this->db->loadResult();
	}

	/**
	 * Set the default variant.
	 *
	 * @param   integer  $productParentId   The product parent ID
	 * @param   integer  $characteristicId  The characteristic ID
	 * @param   string   $alias             The characteristic alias
	 *
	 * @return  void
	 *
	 * @since   7.7.0
	 */
	private function storeDefaultVariant($productParentId, $characteristicId, $alias)
	{
		$this->log->add('Storing default variant for ' . $this->getState($alias, false), false);

		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('variant_characteristic_id'))
			->from($this->db->quoteName('#__hikashop_variant'))
			->where($this->db->quoteName('variant_product_id') . ' = ' . (int) $productParentId)
			->where($this->db->quoteName('variant_characteristic_id') . ' = ' . (int) $characteristicId);
		$this->db->setQuery($query);

		$this->log->add('Check if default variant already exists');

		if (!$this->db->loadResult())
		{
			$query->clear()
				->insert($this->db->quoteName('#__hikashop_variant'))
				->columns(
					$this->db->quoteName(
						array(
							'variant_characteristic_id',
							'variant_product_id'
						)
					)
				)
				->values($characteristicId . ', ' . $productParentId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Setting default variant to ' . $this->getState($alias, false));
		}
	}

	/**
	 * Process the characteristic values.
	 *
	 * @param   integer  $variantProductId         The variant product ID
	 * @param   integer  $characteristicId         The characteristic ID
	 * @param   integer  $variantCharacteristicId  The variant characteristic ID
	 *
	 * @return  void
	 *
	 * @since   7.7.0
	 */
	private function updateVariantCharacteristic($variantProductId, $characteristicId, $variantCharacteristicId)
	{
		$query = $this->db->getQuery(true)
			->update($this->db->quoteName('#__hikashop_variant'))
			->set($this->db->quoteName('variant_characteristic_id') . ' = ' . (int) $characteristicId)
			->where($this->db->quoteName('variant_product_id') . ' = ' . (int) $variantProductId)
			->where($this->db->quoteName('variant_characteristic_id') . ' = ' . (int) $variantCharacteristicId);
		$this->db->setQuery($query)->execute();
		$this->log->add('Update variant characteristic value');
	}

	/**
	 * Insert a new variant characteristic.
	 *
	 * @param   integer  $variantProductId  The variant product ID
	 * @param   integer  $characteristicId  The characteristic ID
	 *
	 * @return  void
	 *
	 * @since   1.0.0
	 */
	private function addVariantCharacteristic($variantProductId, $characteristicId)
	{
		// Find out the next ordering position
		$query = $this->db->getQuery(true)
			->select('MAX(' . $this->db->quoteName('ordering') . ') + 1 AS' . $this->db->quoteName('max'))
			->from($this->db->quoteName('#__hikashop_variant'))
			->where($this->db->quoteName('variant_product_id') . ' = ' . (int) $variantProductId)
			->order($this->db->quoteName('ordering'));
		$this->db->setQuery($query);
		$newOrdering = $this->db->loadResult();
		$this->log->add('Get the new ordering position');

		if (!$newOrdering)
		{
			$newOrdering = 0;
		}

		$query->clear()
			->insert($this->db->quoteName('#__hikashop_variant'))
			->columns(
				$this->db->quoteName(
					array(
						'variant_characteristic_id',
						'variant_product_id',
						'ordering'
					)
				)
			)
			->values($characteristicId . ', ' . $variantProductId . ', ' . $newOrdering);
		$this->db->setQuery($query)->execute();
		$this->log->add('Insert variant characteristic');
	}

	/**
	 * For every parent CSVI inserts a default row in #__hikashop_variants table. Now this is correct for parents with
	 * variants but when there is no child product for a variant these default rows shows as 0 variants on HikaShop
	 * Products page. CSVI wouldn't know if a parent has a child or not unless the import is complete so collecting
	 * all parent products and parent products with child in two different arrays. After the import with this array CSVI
	 * finds out the parent product without variants and delete the defaults so the 0 variants message is removed.
	 *
	 * @return  void
	 *
	 * @since   7.7.0
	 */
	public function onAfterStart()
	{
		$productsWithChild = array_unique($this->childWithParent);
		$removeProducts    = array_diff($this->parentProducts, $productsWithChild);

		// Remove the defaults of parents who dont have variants
		if ($removeProducts)
		{
			$query = $this->db->getQuery(true)
				->delete($this->db->quoteName('#__hikashop_variant'))
				->where($this->db->quoteName('variant_product_id') . ' IN (' . implode(',', $removeProducts) . ')');
			$this->db->setQuery($query)->execute();
			$this->log->add('Remove default parent value for products with no variants');
		}
	}

	/**
	 * Process product options.
	 *
	 * @return  True if everything is Ok False otherwise.
	 *
	 * @since   7.3.0
	 *
	 * @throws  \RuntimeException
	 */
	private function processProductOptions()
	{
		$relations = explode('|', $this->getState('product_options', false));

		if (empty($relations))
		{
			$this->log->addStats('incorrect', 'COM_CSVI_NO_RELATED_PRODUCTS_FOUND');

			return false;
		}

		// Store the new relations
		foreach ($relations as $ordering => $related)
		{
			$options   = new \stdClass;
			$productId = $this->helper->getProductId($related);

			if (!$productId)
			{
				$this->log->add('No product id found with product_code ' . $related . ', product not saved as an option');
			}
			else
			{
				// Build the object to store
				$options->product_id               = $this->getState('product_id', false);
				$options->product_related_type     = 'options';
				$options->product_related_id       = $productId;
				$options->product_related_ordering = $ordering + 1;

				// Bind the data
				$this->productRelatedTable->save($options);

				$this->log->add('Product options saved', false);
				$this->log->addStats('Added', 'COM_CSVI_RELATED_PRODUCTS_OPTIONS_SAVED');

				// Clean the table object for next insert
				$this->productRelatedTable->reset();
			}
		}

		return true;
	}

	/**
	 * Get Warehouse ID based on Warehouse name
	 *
	 * @param   string  $warehouseName  Warehouse name
	 *
	 * @return  int ID of the warehouse
	 *
	 * @since   7.18.0
	 */
	private function getWarehouseId(string $warehouseName)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('warehouse_id'))
			->from($this->db->quoteName('#__hikashop_warehouse'))
			->where($this->db->quoteName('warehouse_name') . ' = ' . $this->db->quote($warehouseName));
		$this->db->setQuery($query);
		$value = $this->db->loadResult();
		$this->log->add(($value) ? 'Found Warehouse ID : ' . $value : 'No Warehouse ID value found');

		return $value ?? '';
	}
}
